#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <usb.h>
#include <pthread.h>

#include "m_pd.h"

#include "../../include/phimatics_cmds.h"  // common for firmware & externals


/* TODO
compile on windows:
//#include "lusb0_usb.h"
//#include "pthread.h"
gcc -O3 -funroll-loops -fomit-frame-pointer -mms-bitfields -L"C:\Program Files (x86)\pd\bin" -L"C:\Users\pcoulombe\Dropbox\phimatics\pd\externals" -lpd -lkernel32 -lwsock32 -luser32 -lgdi32 -lusb32 -lpthreadGC2 -shared phimatics.c -o phimatics.pd_linux
*/


// ==============================================================================
// Constants
// ------------------------------------------------------------------------------
#define USBDEV_SHARED_VENDOR    	0x16c0  /* VOTI */
#define USBDEV_SHARED_PRODUCT   	0x05dc  /* Obdev's free shared PID */
#define USB_CFG_VENDOR_NAME			"workinprogress.ca"
#define USB_CFG_DEVICE_NAME			"phimatics"
#define OUTLETS 					1
#define BUFFERINSIZE 				2
#define BUFFEROUTSIZE 				3

// ==============================================================================
// Our External's Memory structure
// ------------------------------------------------------------------------------
typedef struct _phimatics				// defines our object's internal variables for each instance in a patch
{
	t_object 		p_ob;					// object header - ALL pd external MUST begin with this...
	usb_dev_handle	*dev_handle;			// handle to the phimatics usb device
	void 			*outlets[OUTLETS];		// handle to the objects outlets
	int				x_verbose;
	pthread_attr_t 	phimatics_thread_attr;
	pthread_t    	x_threadid;
} t_phimatics;

void *phimatics_class;					// global pointer to the object class - so pd can reference the object 


// ==============================================================================
// Function Prototypes
// ------------------------------------------------------------------------------
void *phimatics_new(t_symbol *s);
void phimatics_assist(t_phimatics *x, void *b, long m, long a, char *s);			
static int usbGetStringAscii(usb_dev_handle *dev, int ndex, int langid, char *buf, int buflen);
void find_device(t_phimatics *x);

static unsigned char threadOutCmd;
static unsigned char bufferOut[BUFFEROUTSIZE];

// =============================================================================
// Threading
// ------------------------------------------------------------------------------
static void *usb_thread_read(void *w)
{
	int nBytes;
	t_phimatics *x = (t_phimatics*) w;
	unsigned char bufferIn[BUFFERINSIZE];

	while(1) {
		pthread_testcancel();
		if(threadOutCmd) {
			nBytes = usb_control_msg(x->dev_handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_OUT, threadOutCmd, 0, 0, (char *)bufferOut, sizeof(bufferOut), 5000);
			threadOutCmd = 0;
		} else {
			nBytes = usb_control_msg(x->dev_handle, USB_TYPE_VENDOR | USB_RECIP_DEVICE | USB_ENDPOINT_IN, PHIMATICS_CMD_POLL, 0, 0, (char *)bufferIn, sizeof(bufferIn), 5000);
			outlet_float(x->outlets[0], bufferIn[0] << 8 | bufferIn[1]);
		}
	}
	return 0;
}

static void usb_thread_start(t_phimatics *x) {
	// create the worker thread
    if(pthread_attr_init(&x->phimatics_thread_attr) < 0)
	{
       error("phimatics: could not launch receive thread");
       return;
    }
    if(pthread_attr_setdetachstate(&x->phimatics_thread_attr, PTHREAD_CREATE_DETACHED) < 0)
	{
       error("phimatics: could not launch receive thread");
       return;
    }
    if(pthread_create(&x->x_threadid, &x->phimatics_thread_attr, usb_thread_read, x) < 0)
	{
       error("phimatics: could not launch receive thread");
       return;
    }
    else
    {
       //if(x->x_verbose)post("phimatics: thread %d launched", (int)x->x_threadid );
    }
}


//--------------------------------------------------------------------------
// - Message: allOff
//--------------------------------------------------------------------------
void phimatics_allOff(t_phimatics *x)
{
	threadOutCmd = PHIMATICS_CMD_ALLOFF;
}

//--------------------------------------------------------------------------
// - Message: led
//--------------------------------------------------------------------------
void phimatics_allLed(t_phimatics *x, t_floatarg pwm)
{
	// pwm range limit
	if(pwm > 4095) {
		pwm = 4095;
	} else if(pwm < 0) {
		pwm = 0;
	}

	bufferOut[0] = (int)pwm >> 8; // 12 bits pwm
	bufferOut[1] = (int)pwm;
	threadOutCmd = PHIMATICS_CMD_ALLLED;
}


//--------------------------------------------------------------------------
// - LIST: [float channel [4bits], float pwm [12 bits]]
//--------------------------------------------------------------------------
void phimatics_led(t_phimatics *x, t_symbol *s, int argc, t_atom *argv)
{
	if (argc >= 2) {
		int nBytes;

		int pwm = atom_getint(argv);
		int channel = atom_getint(argv+1);

		// pwm range limit
		if(pwm > 4095) {
			pwm = 4095;
		} else if(pwm < 0) {
			pwm = 0;
		}

		// channel range limit
		if(channel > 48) {
			channel = 48;
		} else if(channel < 1) {
			channel = 1;
		}
		
		bufferOut[0] = channel - 1;
		bufferOut[1] = pwm >> 8; // 12 bits pwm
		bufferOut[2] = pwm;
		threadOutCmd = PHIMATICS_CMD_LED;
	}
}


//--------------------------------------------------------------------------
// - Object creation and setup
//--------------------------------------------------------------------------
int phimatics_setup(void)
{
	phimatics_class = class_new ( gensym("phimatics"),(t_newmethod)phimatics_new, 0, sizeof(t_phimatics), 	CLASS_DEFAULT,0);
	
	// Add message handlers
	class_addmethod(phimatics_class, (t_method)phimatics_allOff, gensym("alloff"), A_DEFSYM,0);
	class_addmethod(phimatics_class, (t_method)phimatics_allLed, gensym("allled"), A_DEFFLOAT,0);
	class_addlist (phimatics_class, phimatics_led);
	post("bald-approved phimatics version 0.1", 0);
	return 1;
}


//--------------------------------------------------------------------------
void *phimatics_new(t_symbol *s)		// s = optional argument typed into object box (A_SYM) -- defaults to 0 if no args are typed
{
	t_phimatics *x;									// local variable (pointer to a t_phimatics data structure)
	x = (t_phimatics *)pd_new(phimatics_class);			 // create a new instance of this object
    x->x_verbose = 1;
	x->dev_handle = NULL;
	find_device(x);
	
	int i;
	// create outlets and assign it to our outlet variable in the instance's data structure
	for (i=0; i < OUTLETS; i++) {
		x->outlets[i] = outlet_new(&x->p_ob, &s_float);
	}	

    usb_thread_start(x); //start polling the device

	return x; // return a reference to the object instance 
}


//--------------------------------------------------------------------------
// - Object destruction
//--------------------------------------------------------------------------
void phimatics_free(t_phimatics *x)
{
	if (x->dev_handle) usb_close(x->dev_handle);
		while(pthread_cancel(x->x_threadid) < 0)
			if(x->x_verbose)post("phimatics: killing thread\n");
		if(x->x_verbose)post("phimatics: thread canceled\n");
    		
}


//--------------------------------------------------------------------------
// - USB Utility Functions
//--------------------------------------------------------------------------
static int  usbGetStringAscii(usb_dev_handle *dev, int ndex, int langid, char *buf, int buflen)
{
char    asciibuffer[256];
int     rval, i;

    if((rval = usb_control_msg(dev, USB_ENDPOINT_IN, USB_REQ_GET_DESCRIPTOR, (USB_DT_STRING << 8) + ndex, langid, asciibuffer, sizeof(asciibuffer), 1000)) < 0)
        return rval;
    if(asciibuffer[1] != USB_DT_STRING)
        return 0;
    if((unsigned char)asciibuffer[0] < rval)
        rval = (unsigned char)asciibuffer[0];
    rval /= 2;
    /* lossy conversion to ISO Latin1 */
    for(i=1;i<rval;i++){
        if(i > buflen)  /* destination buffer overflow */
            break;
        buf[i-1] = asciibuffer[2 * i];
        if(asciibuffer[2 * i + 1] != 0)  /* outside of ISO Latin1 range */
            buf[i-1] = '?';
    }
    buf[i-1] = 0;
    return i-1;
}

//--------------------------------------------------------------------------
void find_device(t_phimatics *x) {
	usb_dev_handle      *handle = NULL;
	struct usb_bus      *bus;
	struct usb_device   *dev;
	
	usb_init();
	usb_find_busses();
    usb_find_devices();
	 for(bus=usb_busses; bus; bus=bus->next){
        for(dev=bus->devices; dev; dev=dev->next){
            if(dev->descriptor.idVendor == USBDEV_SHARED_VENDOR && dev->descriptor.idProduct == USBDEV_SHARED_PRODUCT){
                char    string[256];
                int     len;
                handle = usb_open(dev); /* we need to open the device in order to query strings */
                if(!handle){
                    error ("Warning: cannot open USB device: %s", usb_strerror());
                    continue;
                }
                /* now find out whether the device actually is phimatics */
                len = usbGetStringAscii(handle, dev->descriptor.iManufacturer, 0x0409, string, sizeof(string));
                if(len < 0){
                    post("phimatics: warning: cannot query manufacturer for device: %s", usb_strerror());
                    goto skipDevice;
                }
                
                if(strcmp(string, USB_CFG_VENDOR_NAME) != 0)
                    goto skipDevice;
                len = usbGetStringAscii(handle, dev->descriptor.iProduct, 0x0409, string, sizeof(string));
                if(len < 0){
                    post("phimatics: warning: cannot query product for device: %s", usb_strerror());
                    goto skipDevice;
                }
                if(strcmp(string, USB_CFG_DEVICE_NAME) == 0)
                    break;
				skipDevice:
                usb_close(handle);
                handle = NULL;
            }
        }
        if(handle)
            break;
    }
	
    if(!handle){
        post("!!!Could not find USB device %s", USB_CFG_DEVICE_NAME);
		x->dev_handle = NULL;
	} else {
		 x->dev_handle = handle;
		 post("Found USB device %s", USB_CFG_DEVICE_NAME);
	}
}
